#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Author: Ingelrest François (Athropos@gmail.com)
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA

import gobject, gtk, modules

from modules import Module


# The format of a combo box entry
(
    EX_NAME,         # Displayed name
    EX_ICON,         # Displayed icon
    EX_MODULE,       # The module that manages the explorer
    EX_IS_SEPARATOR  # True if the entry is a separator
) = range(4)


class Explorer(Module):
    """
        This module manages the left part that allows the user to browse media files (e.g., file explorer, library explorer)
        It adds/remove entries to the combo box and pages to the notebook as requested by other modules
        The tabs of the notebook are not shown, the switch is automatically done by this module based on the combo box' selection
    """


    def __init__(self, wTree):
        """
            Constructor
        """
        Module.__init__(self)

        self.combo     = wTree.get_widget('combo-explorer')
        self.notebook  = wTree.get_widget('notebook-explorer')
        self.exWidgets = []
        self.exModules = {}

        self.explorers = gtk.ListStore(gobject.TYPE_STRING, gtk.gdk.Pixbuf, gobject.TYPE_PYOBJECT, gobject.TYPE_BOOLEAN)
        self.combo.set_model(self.explorers)
        txtRdr = gtk.CellRendererText()
        imgRdr = gtk.CellRendererPixbuf()
        self.combo.pack_start(imgRdr, True)
        self.combo.add_attribute(imgRdr, 'pixbuf', EX_ICON)
        self.combo.pack_start(txtRdr, True)
        self.combo.add_attribute(txtRdr, 'text', EX_NAME)

        self.nullIcon = gtk.gdk.Pixbuf(gtk.gdk.COLORSPACE_RGB, True, 8, 16, 16)
        self.nullIcon.fill(0x00000000)
        # Messages handled by this module
        modules.register(self, [modules.MSG_CMD_EXPLORER_ADD])
        # GTK handlers
        wTree.signal_connect('onComboExplorerChanged', self.onExplorerChanged)


    def addExplorer(self, module, name, pixbuf, widget):
        """
            Add a new explorer
            Each widget has only one notebook page associated to it
            So, if the widget is already known, no new page is created
        """
        if pixbuf is None:
            pixbuf = self.nullIcon

        # Is it a new widget?
        if widget not in self.exWidgets:
            self.exWidgets.append(widget)
            self.notebook.append_page(widget)

        # Is it a new module?
        if module not in self.exModules:
            self.exModules[module] = [(name, pixbuf)]
        else:
            self.exModules[module].append((name, pixbuf))

        # Update the combo box
        self.combo.freeze_child_notify()
        self.explorers.clear()
        for module, data in self.exModules.items():
            for (exName, exIcon) in data:
                self.explorers.append((exName, exIcon, module, False))
        self.combo.thaw_child_notify()

        self.combo.set_active(0)


   # --== Message handler ==--


    def handleMsg(self, msg, params):
        """
            A message has been received
        """
        if msg == modules.MSG_CMD_EXPLORER_ADD: self.addExplorer(params['module'], params['name'], params['pixbuf'], params['widget'])


    # --== GTK handlers ==--


    def onExplorerChanged(self, combo):
        """
            A new explorer has been selected with the combo box
        """
        iter = self.explorers.get_iter(self.combo.get_active())
        modules.postMsg(modules.MSG_EVT_EXPLORER_CHANGED, {'module': self.explorers.get_value(iter, EX_MODULE), 'name': self.explorers.get_value(iter, EX_NAME)})
