#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Author: Ingelrest François (Athropos@gmail.com)
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA

import gobject, gtk, gtk.glade, media, modules, os, tools

from tools   import consts, playlist
from modules import Module
from gettext import gettext as _


# The format of a row in the treeview
(
    ROW_NAME,      # Item name
    ROW_TYPE,      # The type of the item (e.g., directory, file)
    ROW_PIXBUF,    # Item icon
    ROW_FULLPATH   # The full path to the item
) = range(4)


# The possible types for an item
# NOTE that the order used here defines the order in which items are displayed in the treeview
(
    TYPE_DIR,      # A directory
    TYPE_PLAYLIST, # A playlist
    TYPE_FILE,     # A media file
    TYPE_NONE      # A fake item, used to display a '+' in front of a directory when needed
) = range(4)


# Data associated with a root directory
(
    ROOT_PIXBUF,    # The pixbuf to display in the explorer combo box
    ROOT_FULLPATH   # The full path to the root
) = range(2)


class FileExplorer(Module):
    """
        This explorer lets the user browse the disk from a given root directory (e.g., ~/, /)
        It uses a 'lazy' treeview (i.e., it is populated only when needed)
    """


    def __init__(self, wTree):
        """
            Constructor
        """
        Module.__init__(self)
        self.currRoot        = None
        self.showHiddenFiles = False
        # Widgets
        self.tree         = gtk.TreeView()
        self.popupMenu    = None
        self.scrolledWin  = gtk.ScrolledWindow();
        # Icons
        self.dirIcon      = self.tree.render_icon(gtk.STOCK_DIRECTORY, gtk.ICON_SIZE_MENU)
        self.mediaIcon    = self.tree.render_icon(gtk.STOCK_FILE,      gtk.ICON_SIZE_MENU).copy()  # We need a copy to modify it
        self.dirMediaIcon = self.tree.render_icon(gtk.STOCK_DIRECTORY, gtk.ICON_SIZE_MENU).copy()  # We need a copy to modify it

        cdrom = self.tree.render_icon(gtk.STOCK_CDROM, gtk.ICON_SIZE_MENU)
        cdrom.composite(self.mediaIcon,    5, 5, 11, 11, 5, 5, 0.6875, 0.6875, gtk.gdk.INTERP_HYPER, 255)
        cdrom.composite(self.dirMediaIcon, 5, 5, 11, 11, 5, 5, 0.6875, 0.6875, gtk.gdk.INTERP_HYPER, 255)
        # The available roots (will be eventually chosen by the user)
        self.roots = {_('Home directory'): (self.tree.render_icon(gtk.STOCK_HOME, gtk.ICON_SIZE_MENU),     consts.dirUsr),
                      _('Root directory'): (self.tree.render_icon(gtk.STOCK_HARDDISK, gtk.ICON_SIZE_MENU), '/',         )}
        # Configure the widgets
        self.scrolledWin.add(self.tree)
        self.scrolledWin.set_shadow_type(gtk.SHADOW_IN)
        self.scrolledWin.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        self.scrolledWin.show()

        imgRdr = gtk.CellRendererPixbuf()
        txtRdr = gtk.CellRendererText()
        column = gtk.TreeViewColumn()
        self.tree.append_column(column)
        self.tree.set_headers_visible(False)
        column.pack_start(imgRdr, False)
        column.add_attribute(imgRdr, 'pixbuf', ROW_PIXBUF)
        column.pack_start(txtRdr, False)
        column.add_attribute(txtRdr, 'text', ROW_NAME)
        column.set_sizing(gtk.TREE_VIEW_COLUMN_AUTOSIZE)

        self.store = gtk.TreeStore(gobject.TYPE_STRING, gobject.TYPE_INT, gtk.gdk.Pixbuf, gobject.TYPE_STRING)
        self.tree.set_model(self.store)
        self.tree.show()
        # Messages handled by this module
        modules.register(self, [modules.MSG_EVT_APP_STARTED, modules.MSG_EVT_EXPLORER_CHANGED])
        # GTK handlers
        self.tree.connect('row-expanded',       self.onRowExpanded)
        self.tree.connect('button-press-event', self.onMouse)


    def cmpRows(self, r1, r2):
        """
            This function is used for sorting rows in the treestore
        """
        if   r1[ROW_TYPE] == r2[ROW_TYPE]: return cmp(r1[ROW_NAME].lower(), r2[ROW_NAME].lower())
        else:                              return r1[ROW_TYPE] - r2[ROW_TYPE]


    def updateChildren(self, parentPath, firstChild):
        """
            Update 4 children, starting from firstChild, and then repost the same task if there are some children left
        """
        if parentPath is None: parent = None
        else:                  parent = self.store.get_iter(parentPath)

        self.tree.freeze_child_notify()
        for childId in range(firstChild, firstChild+4):
            # Get the current child, if any
            child = self.store.iter_nth_child(parent, childId)
            if child is None:
                break
            # Only directories need to be updated
            if self.store.get_value(child, ROW_TYPE) == TYPE_DIR:
                hasChild = False
                for filename in [f for f in os.listdir(self.store.get_value(child, ROW_FULLPATH)) if f[0] != '.' or self.showHiddenFiles]:
                    fullpath = os.path.join(self.store.get_value(child, ROW_FULLPATH), filename)
                    if os.path.isdir(fullpath):
                        hasChild = True
                    elif os.path.isfile(fullpath) and (media.isSupported(filename) or playlist.isSupported(filename)):
                        hasChild = True
                        self.store.set_value(child, ROW_PIXBUF, self.dirMediaIcon)
                        break
                if hasChild:
                    self.store.append(child, ('', TYPE_NONE, self.dirIcon, ''))
        self.tree.thaw_child_notify()

        # Continue updating if needed
        if child is not None:
            gobject.idle_add(self.updateChildren, parentPath, firstChild+4)


    def showPopupMenu(self, button, time):
        """
            Show a popup menu
        """
        if self.popupMenu is None:
            wTree          = gtk.glade.XML(os.path.join(consts.dirRes, 'FileExplorerMenu.glade'), domain=consts.appNameShort)
            self.popupMenu = wTree.get_widget('menu-popup')
            # Connect handlers
            wTree.signal_connect('onAdd',  lambda widget: self.playSelection(False))
            wTree.signal_connect('onPlay', lambda widget: self.playSelection(True))
        self.popupMenu.popup(None, None, None, button, time)


    def playSelection(self, replace):
        """
            Replace/extend the tracklist
        """
        model, iter = self.tree.get_selection().get_selected()
        if iter is not None:
            if   self.store.get_value(iter, ROW_TYPE) == TYPE_DIR:      files = tools.getMediaFiles(self.store.get_value(iter, ROW_FULLPATH))
            elif self.store.get_value(iter, ROW_TYPE) == TYPE_FILE:     files = [self.store.get_value(iter, ROW_FULLPATH)]
            elif self.store.get_value(iter, ROW_TYPE) == TYPE_PLAYLIST: files = playlist.load(self.store.get_value(iter, ROW_FULLPATH))

            if files is not None:
                if replace: modules.postMsg(modules.MSG_CMD_TRACKLIST_SET, {'files': files, 'playNow': True, 'sortTracks': True})
                else:       modules.postMsg(modules.MSG_CMD_TRACKLIST_ADD, {'files': files, 'sortTracks': True})


    def exploreDir(self, dir, parent):
        """
            List the content of the given directory and append it to the tree as a child of parent
        """
        rows = []
        for filename in [file for file in os.listdir(dir) if file[0] != '.' or self.showHiddenFiles]:
            fullpath = os.path.join(dir, filename)

            if os.path.isdir(fullpath):
                rows.append((filename, TYPE_DIR, self.dirIcon, fullpath))
            elif os.path.isfile(fullpath):
                if   media.isSupported(filename):    rows.append((filename, TYPE_FILE,     self.mediaIcon,  fullpath))
                elif playlist.isSupported(filename): rows.append((filename, TYPE_PLAYLIST, self.mediaIcon, fullpath))

        if len(rows) != 0:
            rows.sort(self.cmpRows)

            for row in rows:
                self.store.append(parent, row)

            # TinyOS style: since it may take a long time to update all children, we split this task into multiple subtasks
            # We need to provide a path, since iters are transient
            if parent is None: gobject.idle_add(self.updateChildren, None,                        0)
            else:              gobject.idle_add(self.updateChildren, self.store.get_path(parent), 0)


    def onRowExpanded(self, tree, iter, path):
        """
            The expanded row may have a fake child, in which case we need to replace it by the real files/directories
        """
        child = self.store.iter_children(iter)
        if self.store.get_value(child, ROW_TYPE) == TYPE_NONE:
            self.tree.freeze_child_notify()
            self.exploreDir(self.store.get_value(iter, ROW_FULLPATH), iter)
            self.store.remove(child)
            self.tree.thaw_child_notify()


    def onExplorerChanged(self, name):
        """
            A new explorer has been selected, check which one it is
        """
        if name in self.roots and name != self.currRoot:
            self.currRoot = name
            self.tree.set_model(None)
            self.store.clear()
            self.exploreDir(self.roots[name][ROOT_FULLPATH], None)
            self.tree.set_model(self.store)


    # --== GTK handlers ==--


    def onMouse(self, tree, event):
        """
            A mouse button has been pressed
        """
        path = self.tree.get_path_at_pos(int(event.x), int(event.y))
        if path is not None:
            iter = self.store.get_iter(path[0])
            if event.button == 3 and self.store.get_value(iter, ROW_PIXBUF) != self.dirIcon:
                self.showPopupMenu(event.button, event.time)
            elif event.button == 1 and event.type == gtk.gdk._2BUTTON_PRESS:
                if   self.store.get_value(iter, ROW_PIXBUF) != self.dirIcon: self.playSelection(True)
                elif self.tree.row_expanded(path[0]):                        self.tree.collapse_row(path[0])
                else:                                                        self.tree.expand_row(path[0], False)


   # --== Message handler ==--


    def handleMsg(self, msg, params):
        """
            A message has been received
        """
        if msg == modules.MSG_EVT_EXPLORER_CHANGED and params['module'] == self:
            self.onExplorerChanged(params['name'])
        elif msg == modules.MSG_EVT_APP_STARTED:
            for name, (pixbuf, fullpath) in self.roots.items():
                modules.postMsg(modules.MSG_CMD_EXPLORER_ADD, {'module': self, 'name': name, 'pixbuf': pixbuf, 'widget': self.scrolledWin})
