# -*- coding: utf-8 -*-
#
# Author: Ingelrest François (Athropos@gmail.com)
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA

import dbus, gobject, modules

from tools   import consts
from gettext import gettext as _
from modules import ThreadedModule


DEFAULT_STOP_STATUS = _('Decibel is stopped')


##############################################################################


class Gaim:

    def __init__(self, dbusInterface):
        """
            Constructor
        """
        self.dbusInterface = dbusInterface

    def listAccounts(self):
        """
            Gaim merges all accounts, so we return a default one
            Each account is associated with:
                * A boolean -> True if the status of this account was changed on the previous track change
        """
        return {'GenericAccount': False}

    def setAccountStatusMsg(self, account, msg):
        """
            Change the status message of the given account
            Why is it so complex with Gaim??
            Return true if the message is successfully updated
        """
        try:
            current    = self.dbusInterface.GaimSavedstatusGetCurrent()
            statusType = self.dbusInterface.GaimSavedstatusGetType(current)
            statusId   = self.dbusInterface.GaimPrimitiveGetIdFromType(statusType)
            if statusId in ['available']:
                saved = self.dbusInterface.GaimSavedstatusNew('', statusType)
                self.dbusInterface.GaimSavedstatusSetMessage(saved, msg)
                self.dbusInterface.GaimSavedstatusActivate(saved)
                return True
        except:
            pass

        return False


##############################################################################


class Gajim:

    def __init__(self, dbusInterface):
        """
            Constructor
        """
        self.dbusInterface = dbusInterface

    def listAccounts(self):
        """
            Return a list of existing accounts
            Each account is associated with:
                * A boolean -> True if the status of this account was changed on the previous track change
        """
        try:
            return dict([(account, False) for account in self.dbusInterface.list_accounts()])
        except:
            return {}

    def setAccountStatusMsg(self, account, msg):
        """
            Change the status message of the given account
            Return true if the message is successfully updated
        """
        try:
            currentStatus = self.dbusInterface.get_status(account)
            if currentStatus in ['online', 'chat']:
                self.dbusInterface.change_status(currentStatus, msg, account)
                return True
        except:
            pass

        return False


##############################################################################


class Gossip:

    def __init__(self, dbusInterface):
        """
            Constructor
        """
        self.dbusInterface = dbusInterface

    def listAccounts(self):
        """
            Gossip merges all accounts, so we return a default one
            Each account is associated with:
                * A boolean -> True if the status of this account was changed on the previous track change
        """
        return {'GenericAccount': False}

    def setAccountStatusMsg(self, account, msg):
        """
            Change the status message of the given account
            Return true if the message is successfully updated
        """
        try:
            currentStatus, currentMsg = self.dbusInterface.GetPresence('')
            if currentStatus in ['available']:
                self.dbusInterface.SetPresence(currentStatus, msg)
                return True
        except:
            pass

        return False


##############################################################################


class Pidgin:

    def __init__(self, dbusInterface):
        """
            Constructor
        """
        self.dbusInterface = dbusInterface

    def listAccounts(self):
        """
            Pidgin merges all accounts, so we return a default one
            Each account is associated with:
                * A boolean -> True if the status of this account was changed on the previous track change
        """
        return {'GenericAccount': False}

    def setAccountStatusMsg(self, account, msg):
        """
            Change the status message of the given account
            Why is it so complex with Pidgin??
            Return true if the message is successfully updated
        """
        try:
            current    = self.dbusInterface.PurpleSavedstatusGetCurrent()
            statusType = self.dbusInterface.PurpleSavedstatusGetType(current)
            statusId   = self.dbusInterface.PurplePrimitiveGetIdFromType(statusType)
            if statusId in ['available']:
                saved = self.dbusInterface.PurpleSavedstatusNew('', statusType)
                self.dbusInterface.PurpleSavedstatusSetMessage(saved, msg)
                self.dbusInterface.PurpleSavedstatusActivate(saved)
                return True
        except:
            pass

        return False


##############################################################################


# Elements associated with each supported IM clients
(
    IM_DBUS_SERVICE_NAME,
    IM_DBUS_OBJECT_NAME,
    IM_DBUS_INTERFACE_NAME,
    IM_CLASS,
    IM_INSTANCE,
    IM_ACCOUNTS
) = range(6)


# All specific classes have been defined, so we can now populate the list of supported IM clients
mSupportedIMClients = [
    ['net.sf.gaim.GaimService',        '/net/sf/gaim/GaimObject',        'net.sf.gaim.GaimInterface',         Gaim,   None, {}],  # Gaim
    ['org.gajim.dbus',                 '/org/gajim/dbus/RemoteObject',   'org.gajim.dbus.RemoteInterface',    Gajim,  None, {}],  # Gajim
    ['org.gnome.Gossip',               '/org/gnome/Gossip',              'org.gnome.Gossip',                  Gossip, None, {}],  # Gossip
    ['im.pidgin.purple.PurpleService', '/im/pidgin/purple/PurpleObject', 'im.pidgin.purple.PurpleInterfacep', Pidgin, None, {}]   # Pidgin
]


class IMStatus(ThreadedModule):
    """
        This module updates the online status of all supported IM clients
    """


    def __init__(self, wTree):
        """
            Constructor
        """
        ThreadedModule.__init__(self)
        self.status  = DEFAULT_STOP_STATUS
        self.trkInfo = ''
        self.clients = []
        # Messages handled by this module
        modules.register(self, [modules.MSG_EVT_NEW_TRACK, modules.MSG_EVT_PAUSED, modules.MSG_EVT_UNPAUSED, modules.MSG_EVT_STOPPED])
        modules.register(self, [modules.MSG_EVT_APP_STARTED, modules.MSG_EVT_APP_QUIT])


    def setStatusMsg(self, status):
        """
            Try to update the status of all accounts of all active IM clients
        """
        self.status = status
        for client in self.clients:
            for account, wasUpdated in client[IM_ACCOUNTS].iteritems():
                client[IM_ACCOUNTS][account] = client[IM_INSTANCE].setAccountStatusMsg(account, status)


    def newTrack(self, trackInfo):
        """
            A new track is being played
        """
        self.trkInfo = '♫ %s - %s ♫' % (trackInfo[consts.TI_ARTIST], trackInfo[consts.TI_ALBUM])
        if self.trkInfo != self.status:
            self.setStatusMsg(self.trkInfo)


    def detectActiveClients(self):
        """
            Find supported active IM clients
        """
        # Seems to randomly crash when not executed in the GTK main loop?
        activeServices = dbus.SessionBus().get_object('org.freedesktop.DBus', '/org/freedesktop/DBus').ListNames()
        for client in mSupportedIMClients:
            if client[IM_DBUS_SERVICE_NAME] in activeServices:
                obj       = dbus.SessionBus().get_object(client[IM_DBUS_SERVICE_NAME], client[IM_DBUS_OBJECT_NAME])
                interface = dbus.Interface(obj, client[IM_DBUS_INTERFACE_NAME])

                client[IM_INSTANCE] = client[IM_CLASS](interface)
                client[IM_ACCOUNTS] = client[IM_INSTANCE].listAccounts()

                self.clients.append(client)


    # --== Message handler ==--


    def handleMsg(self, msg, params):
        """
            A message has been received
        """
        if   msg == modules.MSG_EVT_PAUSED:      self.setStatusMsg(self.trkInfo + _(' [paused]'))
        elif msg == modules.MSG_EVT_STOPPED:     self.setStatusMsg(DEFAULT_STOP_STATUS)
        elif msg == modules.MSG_EVT_APP_QUIT:    self.setStatusMsg(DEFAULT_STOP_STATUS)
        elif msg == modules.MSG_EVT_UNPAUSED:    self.setStatusMsg(self.trkInfo)
        elif msg == modules.MSG_EVT_NEW_TRACK:   self.newTrack(params['trackInfo'])
        elif msg == modules.MSG_EVT_APP_STARTED: gobject.idle_add(self.detectActiveClients)
