# -*- coding: utf-8 -*-
#
# Author: Ingelrest François (Athropos@gmail.com)
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA

import gobject, gtk, media, modules, os.path, tools

from gui     import fileChooser
from tools   import consts, playlist
from gettext import gettext as _
from modules import Module


(
    ROW_FIL,  # Full path to the file
    ROW_NUM,  # Track number
    ROW_TIT,  # Track title
    ROW_ART,  # Track artist
    ROW_ALB,  # Track album
    ROW_LEN,  # Track length in seconds
    ROW_ICO   # Icon drawn in front of the row
) = range(7)


class Tracklist(Module):
    """
        This module manages the tracklist
    """


    def __init__(self, wTree):
        """
            Constructor
        """
        Module.__init__(self)
        self.index    = None
        self.playtime = 0
        # Retrieve widgets
        self.tree      = wTree.get_widget('tree-tracklist')
        self.window    = wTree.get_widget('win-main')
        self.btnClear  = wTree.get_widget('btn-tracklistClear')
        self.popupMenu = None
        self.tracklist = gtk.ListStore(gobject.TYPE_STRING, gobject.TYPE_INT, gobject.TYPE_STRING, gobject.TYPE_STRING,
                                       gobject.TYPE_STRING, gobject.TYPE_INT, gtk.gdk.Pixbuf)
        self.tree.set_model(self.tracklist)
        self.btnClear.set_sensitive(False)
        # The renderers used for the tree
        txtRightRdr = gtk.CellRendererText()
        txtLeftRdr  = gtk.CellRendererText()
        txtRightRdr.set_property('xalign', 1.0)
        # The icons used in the tree
        self.nullIcon  = gtk.gdk.Pixbuf(gtk.gdk.COLORSPACE_RGB, True, 8, 16, 16)
        self.playIcon  = self.tree.render_icon(gtk.STOCK_MEDIA_PLAY,  gtk.ICON_SIZE_MENU)
        self.pauseIcon = self.tree.render_icon(gtk.STOCK_MEDIA_PAUSE, gtk.ICON_SIZE_MENU)
        self.nullIcon.fill(0x00000000)
        # Create the columns of the tree
        self.tree.insert_column_with_attributes(-1, '#', gtk.CellRendererPixbuf(), pixbuf=ROW_ICO)
        self.tree.get_column(0).pack_start(txtRightRdr, True)
        self.tree.get_column(0).add_attribute(txtRightRdr, 'text', ROW_NUM)
        self.tree.get_column(0).set_sizing(gtk.TREE_VIEW_COLUMN_AUTOSIZE)

        self.tree.insert_column_with_attributes(-1, _('Title'),  txtLeftRdr, text=ROW_TIT)
        self.tree.get_column(1).set_sizing(gtk.TREE_VIEW_COLUMN_AUTOSIZE)
        self.tree.get_column(1).set_expand(True)

        self.tree.insert_column_with_attributes(-1, _('Artist'), txtLeftRdr, text=ROW_ART)
        self.tree.get_column(2).set_sizing(gtk.TREE_VIEW_COLUMN_AUTOSIZE)
        self.tree.get_column(2).set_expand(True)

        self.tree.insert_column_with_attributes(-1, _('Album'),  txtLeftRdr, text=ROW_ALB)
        self.tree.get_column(3).set_sizing(gtk.TREE_VIEW_COLUMN_AUTOSIZE)
        self.tree.get_column(3).set_expand(True)

        self.tree.insert_column_with_attributes(-1, _('Length'), txtRightRdr, text=ROW_LEN)
        self.tree.get_column(4).set_sizing(gtk.TREE_VIEW_COLUMN_AUTOSIZE)
        self.tree.get_column(4).set_cell_data_func(txtRightRdr, lambda c, cll, mdl, it: cll.set_property('text', tools.sec2str(mdl.get_value(it, ROW_LEN))))
        self.tree.enable_model_drag_source(gtk.gdk.BUTTON1_MASK, [("example", 0, 0)], gtk.gdk.ACTION_DEFAULT)
        self.tree.enable_model_drag_dest([("example", 0, 0)], gtk.gdk.ACTION_DEFAULT)
        # Drag'n'drop
        targets = [('TRACKLIST_ROW', gtk.TARGET_SAME_WIDGET, 0)]
        self.tree.enable_model_drag_source(gtk.gdk.BUTTON1_MASK, targets, gtk.gdk.ACTION_DEFAULT | gtk.gdk.ACTION_MOVE)
        self.tree.enable_model_drag_dest(targets, gtk.gdk.ACTION_DEFAULT)
        # Messages handled by this module
        modules.register(self, [modules.MSG_CMD_TRACKLIST_SET, modules.MSG_CMD_TRACKLIST_CLEAR, modules.MSG_CMD_TRACKLIST_ADD,
                                modules.MSG_CMD_TOGGLE_PAUSE, modules.MSG_CMD_NEXT, modules.MSG_CMD_PREVIOUS, modules.MSG_EVT_NEED_BUFFER,
                                modules.MSG_EVT_STOPPED, modules.MSG_EVT_PAUSED, modules.MSG_EVT_UNPAUSED, modules.MSG_EVT_TRACK_ENDED])
        # GTK handlers
        wTree.signal_connect('onTracklistClear',            lambda widget: modules.postMsg(modules.MSG_CMD_TRACKLIST_CLEAR))
        wTree.signal_connect('onTracklistKeyboard',         self.onKeyboard)
        wTree.signal_connect('onTracklistMouseBtn',         self.onMouseButton)
        wTree.signal_connect('onTracklistDragDataGet',      self.onDragDataGet)
        wTree.signal_connect('onTracklistDragDataReceived', self.onDragDataReceived)


    def hasNext(self):      return self.index is not None and self.index < len(self.tracklist) - 1
    def hasPrev(self):      return self.index is not None and self.index > 0
    def getAllFiles(self):  return [row[ROW_FIL] for row in self.tracklist]
    def getTrackInfo(self): return self.tracklist.get(self.tracklist.get_iter(self.index), 0, 1, 2, 3, 4, 5)


    def setVisible(self, widget, show):
        """
            Show/hide a widget
        """
        if show: widget.show()
        else:    widget.hide()


    def cmpTracks(self, t1, t2):
        """
            Used to sort tracks
        """
        if t1[ROW_NUM] != t2[ROW_NUM]: return t1[ROW_NUM] - t2[ROW_NUM]
        else:                          return cmp(t1[ROW_TIT].lower(), t2[ROW_TIT].lower())


    def jumpTo(self, index):
        """
            Jump to the track located at the given index
        """
        if self.index is not None:
            self.tracklist.set_value(self.tracklist.get_iter(self.index), ROW_ICO, self.nullIcon)
        self.index = index
        self.tracklist.set_value(self.tracklist.get_iter(self.index), ROW_ICO, self.playIcon)
        modules.postMsg(modules.MSG_CMD_PLAY, {'filename': self.tracklist.get_value(self.tracklist.get_iter(self.index), ROW_FIL)})
        modules.postMsg(modules.MSG_EVT_NEW_TRACK, {'trackInfo': self.getTrackInfo()})
        modules.postMsg(modules.MSG_EVT_PREVNEXT, {'prevOk': self.hasPrev(), 'nextOk': self.hasNext()})


    def onTrackEnded(self):
        """
            The current track has ended, jump to the next one if any
        """
        if self.hasNext(): self.jumpTo(self.index+1)
        else:              modules.postMsg(modules.MSG_CMD_STOP)


    def onBufferingNeeded(self):
        """
            The current track is close to its end, se we try to buffer the next one to avoid gaps
        """
        if self.hasNext():
            modules.postMsg(modules.MSG_CMD_BUFFER, {'filename': self.tracklist.get_value(self.tracklist.get_iter(self.index+1), ROW_FIL)})


    def append(self, files, sortTracks):
        """
            Append files to the tracklist
        """
        tracks = media.readInfo(files)
        if sortTracks:
            tracks.sort(self.cmpTracks)
        self.tree.freeze_child_notify()
        for track in tracks:
            track.append(self.nullIcon)
            self.tracklist.append(track)
            self.playtime += track[ROW_LEN]
        self.tree.thaw_child_notify()
        self.btnClear.set_sensitive(True)
        modules.postMsg(modules.MSG_EVT_NEW_TRACKLIST, {'files': self.getAllFiles(), 'playtime': self.playtime})
        modules.postMsg(modules.MSG_EVT_PREVNEXT, {'prevOk': self.hasPrev(), 'nextOk': self.hasNext()})


    def set(self, files, playNow, sortTracks):
        """
            Replace the tracklist, clear it if files is None
        """
        self.playtime = 0
        self.index    = None
        self.tracklist.clear()

        if files is None:
            modules.postMsg(modules.MSG_CMD_STOP)
            self.btnClear.set_sensitive(False)
            modules.postMsg(modules.MSG_EVT_NEW_TRACKLIST, {'files': [], 'playtime': 0})
        else:
            self.append(files, sortTracks)
            if playNow:
                self.jumpTo(0)


    def onStopped(self):
        """
            Playback has been stopped
        """
        if self.index is not None:
            self.tracklist.set_value(self.tracklist.get_iter(self.index), ROW_ICO, self.nullIcon)
            self.index = None


    def onPausedToggled(self, icon):
        """
            Switch between paused and unpaused
        """
        if self.index is not None:
            self.tracklist.set_value(self.tracklist.get_iter(self.index), ROW_ICO, icon)


    def playSelected(self):
        """
            Start playing the currently selected track
        """
        model, iter = self.tree.get_selection().get_selected()
        if iter is not None:
            self.jumpTo(self.tracklist.get_path(iter)[0])


    def savePlaylist(self):
        """
            The current tracklist to a playlist
        """
        filename = fileChooser.save(self.window, _('Save playlist'), 'playlist.m3u')
        if filename is not None:
            playlist.save(self.getAllFiles(), filename, False)


    def removeSelected(self):
        """
            Remove the currently selected track (allowed only if it is not being played)
        """
        model, iter = self.tree.get_selection().get_selected()
        if iter is not None and self.tracklist.get_value(iter, ROW_ICO) == self.nullIcon:

            if self.index is not None and self.tracklist.get_path(iter)[0] < self.index:
                self.index -= 1

            self.playtime -= self.tracklist.get_value(iter, ROW_LEN)

            if   self.tracklist.remove(iter): self.tree.set_cursor(self.tracklist.get_path(iter))
            elif len(self.tracklist) == 0:    self.btnClear.set_sensitive(False)

            modules.postMsg(modules.MSG_EVT_PREVNEXT,      {'prevOk': self.hasPrev(),     'nextOk':   self.hasNext()})
            modules.postMsg(modules.MSG_EVT_NEW_TRACKLIST, {'files':  self.getAllFiles(), 'playtime': self.playtime})


    def showPopupMenu(self, index, button, time):
        """
            The index parameter may be None
        """
        if self.popupMenu is None:
            wTree              = gtk.glade.XML(os.path.join(consts.dirRes, 'TracklistMenu.glade'), domain=consts.appNameShort)
            self.menuPlay      = wTree.get_widget('item-play')
            self.menuStop      = wTree.get_widget('item-stop')
            self.popupMenu     = wTree.get_widget('menu-popup')
            self.menuPause     = wTree.get_widget('item-pause')
            self.menuRemove    = wTree.get_widget('item-remove')
            self.menuUnpause   = wTree.get_widget('item-unpause')
            self.menuSeparator = wTree.get_widget('item-separator')
            # Connect signals
            wTree.signal_connect('onSave',    lambda widget: self.savePlaylist())
            wTree.signal_connect('onPlay',    lambda widget: self.playSelected())
            wTree.signal_connect('onStop',    lambda widget: modules.postMsg(modules.MSG_CMD_STOP))
            wTree.signal_connect('onPause',   lambda widget: modules.postMsg(modules.MSG_CMD_TOGGLE_PAUSE))
            wTree.signal_connect('onClear',   lambda widget: modules.postMsg(modules.MSG_CMD_TRACKLIST_CLEAR))
            wTree.signal_connect('onRemove',  lambda widget: self.removeSelected())
            wTree.signal_connect('onUnpause', lambda widget: modules.postMsg(modules.MSG_CMD_TOGGLE_PAUSE))

        # Show only relevant menu entries
        if index is None: icon = None
        else:             icon = self.tracklist.get_value(self.tracklist.get_iter(index), ROW_ICO)

        self.setVisible(self.menuPlay,      icon == self.nullIcon)
        self.setVisible(self.menuStop,      icon != self.nullIcon)
        self.setVisible(self.menuPause,     icon == self.playIcon)
        self.setVisible(self.menuRemove,    icon == self.nullIcon)
        self.setVisible(self.menuUnpause,   icon == self.pauseIcon)
        self.setVisible(self.menuSeparator, icon is not None)

        self.popupMenu.popup(None, None, None, button, time)


    # --== Message handler ==--


    def handleMsg(self, msg, params):
        """
            A message has been received
        """
        if   msg == modules.MSG_EVT_PAUSED:                              self.onPausedToggled(self.pauseIcon)
        elif msg == modules.MSG_EVT_STOPPED:                             self.onStopped()
        elif msg == modules.MSG_EVT_UNPAUSED:                            self.onPausedToggled(self.playIcon)
        elif msg == modules.MSG_EVT_TRACK_ENDED:                         self.onTrackEnded()
        elif msg == modules.MSG_EVT_NEED_BUFFER:                         self.onBufferingNeeded()
        elif msg == modules.MSG_CMD_TRACKLIST_SET:                       self.set(params['files'], params['playNow'], params['sortTracks'])
        elif msg == modules.MSG_CMD_TRACKLIST_ADD:                       self.append(params['files'], params['sortTracks'])
        elif msg == modules.MSG_CMD_TRACKLIST_CLEAR:                     self.set(None, False, False)
        elif msg == modules.MSG_CMD_NEXT and self.hasNext():             self.jumpTo(self.index+1)
        elif msg == modules.MSG_CMD_PREVIOUS and self.hasPrev():         self.jumpTo(self.index-1)
        elif msg == modules.MSG_CMD_TOGGLE_PAUSE and self.index is None: self.jumpTo(0)


    # --== GTK handlers ==--


    def onMouseButton(self, tree, event) :
        """
            Play the selected track on double click, or show a popup menu on right click
        """
        path = self.tree.get_path_at_pos(int(event.x), int(event.y))
        if event.button == 1 and event.type == gtk.gdk._2BUTTON_PRESS and path is not None:
            self.jumpTo(path[0][0])
        elif event.button == 3:
            if   path is not None:         self.showPopupMenu(path[0][0], event.button, event.time)
            elif len(self.tracklist) != 0: self.showPopupMenu(None,       event.button, event.time)


    def onKeyboard(self, tree, event):
        """
            Remove the selection if possible
        """
        if gtk.gdk.keyval_name(event.keyval) == 'Delete':
            self.removeSelected()


    def onDragDataGet(self, tree, context, selection, targetId, etime):
        """
            Drag'n'drop has started
        """
        model, iter = tree.get_selection().get_selected()
        index       = model.get_path(iter)[0]
        selection.set(selection.target, 8, str(index))


    def onDragDataReceived(self, tree, context, x, y, selection, info, etime):
        """
            Drag'n'drop has ended
        """
        idxSrc  = int(selection.data)
        iterSrc = self.tracklist.get_iter(idxSrc)
        row     = self.tracklist.get(iterSrc, 0, 1, 2, 3, 4, 5, 6)
        # Copy the row to the new position
        dropInfo = tree.get_dest_row_at_pos(x, y)
        if dropInfo is not None:
            path, pos  = dropInfo
            idxDst     = path[0]
            iterDst    = self.tracklist.get_iter(path)
            newRowIter = self.tracklist.insert_before(iterDst, row)
        else:
            idxDst     = len(self.tracklist)
            newRowIter = self.tracklist.append(row)
        # Remove source row
        self.tracklist.remove(iterSrc)
        # Update the current index position if needed
        if self.index is not None:
            if   self.index == idxSrc:                         self.index  = self.tracklist.get_path(newRowIter)[0]
            elif idxSrc < self.index and idxDst >  self.index: self.index -= 1
            elif idxSrc > self.index and idxDst <= self.index: self.index += 1
        # The tracklist has changed
        modules.postMsg(modules.MSG_EVT_NEW_TRACKLIST, {'files': self.getAllFiles(), 'playtime': self.playtime})
        modules.postMsg(modules.MSG_EVT_PREVNEXT, {'prevOk': self.hasPrev(), 'nextOk': self.hasNext()})
        context.finish(True, False, etime)
