# -*- coding: utf-8 -*-
#
# Author: Ingelrest François (Athropos@gmail.com)
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA

import gobject, gtk, os, sys, traceback

from tools     import prefsManager
from gettext   import gettext as _
from threading import Lock, Semaphore, Thread


# List of messages that can be sent/received by modules
# A message is always associated with a dictionnary that contains its parameters
# The dictionnary may be empty if there's no parameter
# When nothing is specified, there's no parameter
(
    # Commands: may be sent when an action is requested
    MSG_CMD_PLAY,             # Play a file -- filename
    MSG_CMD_STOP,             # Stop playing the current track, if any
    MSG_CMD_SEEK,             # Jump to a position -- seconds
    MSG_CMD_NEXT,             # Jump to the next track
    MSG_CMD_BUFFER,           # Buffer a file -- filename
    MSG_CMD_PREVIOUS,         # Jump to the previous track
    MSG_CMD_TOGGLE_PAUSE,     # Switch between play/pause
    MSG_CMD_EXPLORER_ADD,     # Add a new explorer -- module, name, pixbuf, widget
    MSG_CMD_TRACKLIST_SET,    # Replace the tracklist -- files, playNow, sortTracks
    MSG_CMD_TRACKLIST_ADD,    # Append files to the tracklist -- files, sortTracks
    MSG_CMD_TRACKLIST_CLEAR,  # Clear the tracklist

    # Events: sent whenever an event occurs
    MSG_EVT_PAUSED,           # The current track has been paused
    MSG_EVT_STOPPED,          # The current track has been stopped
    MSG_EVT_UNPAUSED,         # The current track has been unpaused
    MSG_EVT_PREVNEXT,         # Possible/impossible to perform previous/next -- prevOk, nextOk
    MSG_EVT_APP_QUIT,         # The application is quitting
    MSG_EVT_NEW_TRACK,        # The current track has changed -- trackInfo
    MSG_EVT_NEED_BUFFER,      # The current track is close to its end and the next file should be buffered, if possible
    MSG_EVT_TRACK_ENDED,      # The current track has ended
    MSG_EVT_APP_STARTED,      # The application has just started
    MSG_EVT_NEW_TRACKLIST,    # A new tracklist has been set -- files, playtime
    MSG_EVT_TRACK_POSITION,   # The track position has changed -- seconds
    MSG_EVT_EXPLORER_CHANGED, # A new explorer has been selected -- module, name

    # End value
    MSG_END_VALUE
) = range(24)


mMutex    = Lock()                                            # This mutex protects all the functions from concurrent access
mModules  = {}                                                # All known modules, together with an 'active' boolean
mHandlers = dict([(msg, []) for msg in range(MSG_END_VALUE)]) # For each message, store the list of registered modules


def load(wTree):
    """
        Dynamically load all modules
        The 'wTree' parameter is the tree of widgets returned by Glade
    """
    # Make sure the path is known to the loader
    modDir = os.path.dirname(__file__)
    if not modDir in sys.path:
        sys.path.append(modDir)
    # Find all modules
    for filename in os.listdir(modDir):
        if filename.endswith('.py') and filename != '__init__.py':
            try:
                modName = filename.rsplit('.', 1)[0]
                mod     = __import__(modName)
                mModules[getattr(mod, modName)(wTree)] = True
            except:
                print _('Unable to load module "%s"!') % (os.path.join(modDir, filename))
                traceback.print_exc()
    # Start the threaded modules, this has no effect on non-threaded ones
    for module in mModules:
        module.start()


def register(module, msgList):
    """
        Register the given module for all messages in the given list
        If a module is already registered for a message, nothing happens
    """
    mMutex.acquire()
    for msg in msgList:
        if module not in mHandlers[msg]:
            mHandlers[msg].append(module)
    mMutex.release()


def unload(module):
    """
        Unload the given module if it is active
        Nothing happens if it not active
    """
    mMutex.acquire()
    if mModules[module]:
        for handlers in mHandlers.values():
            if module in handlers:
                handlers.remove(module)
        mModules[module] = False
    mMutex.release()


def __postMsg(msg, params={}):
    """
        This is the 'real' postMsg() function
        It MUST be executed in the main GTK loop
    """
    mMutex.acquire()
    for module in mHandlers[msg]:
        module.postMsg(msg, params)
    mMutex.release()


def __postQuitMsg():
    """
        This is the 'real' postQuitMsg() function
        It MUST be executed in the main GTK loop
    """
    __postMsg(MSG_EVT_APP_QUIT)

    for module in mModules:
        module.join()

    prefsManager.save()
    gtk.main_quit()


def postMsg(msg, params={}):
    """
        Post a message in the queue of all modules
        Only registered modules for this message will receive it
    """
    # We need to ensure that posting messages will be done by the main GTK thread
    # Otherwise, the code of non-threaded modules will be executed in the caller's thread, which may not be the main GTK thread
    # This can cause problems when such modules call GTK functions
    gobject.idle_add(__postMsg, msg, params)


def postQuitMsg():
    """
        Post a MSG_EVT_APP_QUIT in each module's queue and exit the application
        Preferences are saved before exiting
    """
    # As with postMsg(), we need to ensure that the code will be executed by the main GTK thread
    gobject.idle_add(__postQuitMsg)


    # --== Base classes for the modules ==--


class ModuleBase:
    """
        This is the base class for all kinds of modules
    """
    def join(self):                    pass
    def start(self):                   pass
    def __init__(self):                pass
    def postMsg(self, msg, params):    pass
    def handleMsg(self, msg, params):  pass


class Module(ModuleBase):
    """
        This is the base class for all non-threaded modules
    """
    def __init__(self):             ModuleBase.__init__(self)
    def postMsg(self, msg, params): self.handleMsg(msg, params)


class ThreadedModule(Thread, ModuleBase):
    """
        This is the base class for all threaded modules
    """


    def __init__(self):
        """
            Constructor
        """
        Thread.__init__(self)
        ModuleBase.__init__(self)

        self.queue     = []              # List of queued messages
        self.mutex     = Lock()          # Protect access to the queue
        self.semaphore = Semaphore(0)    # Block the thread until some messages are available

        # Some messages required by this class
        register(self, [MSG_EVT_APP_QUIT])


    def postMsg(self, msg, params):
        """
            Enqueue a new message in this threads's message queue
            This function is thread-safe
        """
        self.mutex.acquire()
        self.queue.append((msg, params))
        self.mutex.release()
        self.semaphore.release()


    def run(self):
        """
            The thread's entry point
            The execution is suspended until some messages are available
            When this happens, the main message handler is called
        """
        while True:
            self.semaphore.acquire()
            self.mutex.acquire()
            (msg, params) = self.queue.pop(0)
            self.mutex.release()
            self.handleMsg(msg, params)
            # The thread may need to stop
            if msg == MSG_EVT_APP_QUIT:
                break
